"""Constants and data structures for the pipe abstraction."""
__all__ = [
    "TYPE_UDP_CON",
    "TYPE_UDP_SERVER",
    "TYPE_TCP_CON",
    "TYPE_TCP_SERVER",
    "TYPE_TCP_CLIENT",
    "aionetiface_fds",
]

TYPE_UDP_CON = 1
TYPE_UDP_SERVER = 2
TYPE_TCP_CON = 3
TYPE_TCP_SERVER = 4
TYPE_TCP_CLIENT = 5
aionetiface_fds = set()
